/*GPL*START*
 * 
 * Copyright (C) 1998 by Johannes Overmann <overmann@iname.com>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * *GPL*END*/  

#include "asm.h"
#include "terror.h"  // for fatalError()

#include <ctype.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>


#define EOL 65535L
#define ERR 65534L
#define E_FRef  2000000000
#define E_LocFRef 2000000001
#define E_syntax 2000000002



#define ErrorRef() if(l!=E_syntax) Error(line,"can't resolv reference '%s'",par)

typedef unsigned char uchar;
typedef unsigned short ushort;
typedef unsigned int uint;

typedef struct {
   int x;
   ushort lin;
   bool use;
   char n[max_lablen]; 
} LABEL;

typedef struct {
   ushort lin;
   ushort info;
   ushort adr;
   char e[max_explen]; 
} FREF;

enum {E_branch,E_word,E_byte,E_ubyte,E_adr};

char lab[1024];
char opc[1024];
char par[1024];
char buf[1024];
char line_buf[1024];

LABEL label[max_lab];
FREF  fref[max_fref];
uchar mem[max_mem];

int pmaxlablen=0,maxnumfref=0,maxnumlab=0,numlab=0,numfref=0;
int bline,line=0,warn=0,err=0,max_err=0x7ffffff;
int base,amem;
ushort type;
int terr;

void UpperStr(char *s)
{
 while(*s)
 {
  if((*s>='a')&&(*s<='z'))*s-=32;
  s++;
 }
}


void Warn(ushort l,char *fmt, ...) {
   va_list ap;

   warn++;
   if(!warnings_enabled) return;
   va_start(ap,fmt);
   printf("(warning (line %5d): ",l);
   vprintf(fmt,ap);
   printf(")\n");
   if(line == l) {      
      printf("   [%s]\n", line_buf);
   }
   va_end(ap);
}


void Error(ushort l,char *fmt, ...) {
   va_list ap;
   
   err++;
   if(err>=max_err) return;
   va_start(ap,fmt);
   printf("error    (line %5d): ",l);
   vprintf(fmt,ap);
   printf("!\n");
   if(line == l) {      
      printf("   [%s]\n", line_buf);
   }
   va_end(ap);
}

ushort LenStr(char *s)
{
 ushort i=0;

 while(*s++)i++;
 return(i);
}

void CopyStr(char *s, char *d)
{
 do
 {
  *d++=*s;
 }while(*s++);
}

bool CharInStr(char *s,char a)
{
 while(*s)
 {
  if(*s==a)return true;
  s++;
 }
 return false;
}


// this is in a mess:
bool GetLine(void) {
   short i=0,j;

   do {
      if(fgets(buf, sizeof(buf)-1, ifile)==0) return false;
      if(verbose) printf("reading line %4d\r", ++line);
   } while(buf[0]=='*');
   buf[sizeof(buf)-1]=0;
   strcpy(line_buf, buf);
   for(j=0; line_buf[j]; j++) 
     if(line_buf[j]=='\n') 
       line_buf[j]=0;
   for(j=0;buf[j];j++) if(buf[j]==';') buf[j]=0;
   j=0;
   while(buf[i]>' ')lab[j++]=buf[i++];
   lab[j]=0;
   while((buf[i]<=' ')&&(buf[i]!=0))i++;
   j=0;
   while(buf[i]>' ')opc[j++]=buf[i++];
   opc[j]=0;
   while((buf[i]<=' ')&&(buf[i]!=0))i++;
   j=0;
   while(buf[i]>' ')par[j++]=buf[i++];
   par[j]=0;
   UpperStr(opc);
   return true;
}


// just for convenience (I did not know about tstring.h very well :)
bool SameStr(char *a,char *b) {
   return strcmp(a,b)==0;
}


ushort SearchLabel(char *s)
{
 ushort i;

 s[max_lablen-1]=0;
 for(i=0;i<numlab;i++)
 {
  if(SameStr(s,label[i].n))
  {
   return(i);
  }
 }
 return(EOL);
}


void PutMem(ushort a,uchar b) {
   a-=base;
   if(a>=max_mem) {
      Error(line,"insufficient object code memory");
      amem=base;
   } else {
      mem[a]=b;
   }
}


void CheckRange1(ushort li,int a,int b,int l) {
   if((l<a)||(l>b)) {
      Error(li,"(line %d) value out of range [%d..%d] (was %d)",line,a,b,l);
   }
}


void CheckRange(int a,int b,int *l) {
   if((*l<a)||(*l>b)) {
      Error(line,"value out of range [%d..%d] (was %d)",a,b,*l);
      *l=a;
   }
}



int NumEvalP(char **p,char *s,int base) {
   int l=0;
   uchar c;

   if(**p=='$') {
      (*p)++;
      return(NumEvalP(p,"01234567",8L));
   }
   if(!CharInStr(s,**p))return(E_syntax);
   while(CharInStr(s,**p)) {
      c=*((*p)++);
      l=l*base+(c-'0');
      if(c>='a') 
	l-=7+32;
      else if(c>='A') 
	l-=7;
   }
   return(l);
}


int NumEval(char **p)
{
 switch(*((*p)++))
 {
  case '$':return(NumEvalP(p,"0123456789ABCDEFabcdef",16L));
  case '&':return(NumEvalP(p,"01234567",8L));
  case '%':return(NumEvalP(p,"01",2L));
  case '=':return(NumEvalP(p,"0123",4L));
  case '\'':return(*((*p)++));
  default:
   (*p)--;
   return(NumEvalP(p,"0123456789",10L));
 }
}


int LabEval(char **p)
{
 ushort i;
 char *a="0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ._";
 char *d=buf;

 if(!CharInStr(a,**p))return(E_syntax);
 while(CharInStr(a,**p))
 {
  *(d++)=*((*p)++);
 }
 *d=0;
 i=SearchLabel(buf);
 if(i==EOL)
 {
  if(*buf=='.')return(E_LocFRef);
  else return(E_FRef);
 }
 else
 {
  label[i].use=true;
  return(label[i].x);
 }
}

int EvalP(char *p1)
{
 char *p=p1;
 int l;
 char *tnum="$%'=&0123456789";

 type=E_word;
 if(*p==0)
 {
  Error(line,"expecting expression");
  terr=E_syntax;
  return(E_syntax);
 }
 if(CharInStr(tnum,*p))
 {
  l=NumEval(&p);
  if(l>=E_syntax-2)
  {
   Error(line,"expecting value");
   terr=E_syntax;
   return(l);
  }
 }
 else
 {
  switch(*p)
  {
   case '-':
    p++;
    return(-EvalP(p));
   case '+':
    p++;
    return(+EvalP(p));
   case '~':
    p++;
    return(~EvalP(p));
   default:
    l=LabEval(&p);
    if(l==E_syntax)
    {
     Error(line,"illegal char in expression");
     terr=E_syntax;
     return(l);
    }
  }
 }
 if(l>=E_FRef)
 {
  if(terr!=E_syntax)
  {
   if(terr!=E_LocFRef)terr=l;
  }
  l=65535L;
 }
 switch(*p++)
 {
  case 0: return(l);
  case '+':return(l+EvalP(p));
  case '-':return(l-EvalP(p));
  case '*':return(l*EvalP(p));
  case '/':return(l/EvalP(p));
  case '%':return(l%EvalP(p));
  case '&':return(l&EvalP(p));
  case '|':return(l|EvalP(p));
  case '^':return(l^EvalP(p));
  case '>':return(l>>EvalP(p));
  case '<':return(l<<EvalP(p));
  default:
   Error(line,"expression syntax in '%s'",--p);
   terr=E_syntax;
   return(E_syntax);
 }
}


int Eval(char *p) {
   int l;
   
   type=E_word;
   terr=0L;
   l=EvalP(p);
   if((l>=0L)&&(l<256L)) {
      type = E_byte;
      if((*p=='$') && (strlen(p)==5)) type = E_word;
   }
   if(terr>0)return(terr);
   else return(l);
}


int FEval(ushort a)
{
 int l;

 l=Eval(par);
 if(l<E_FRef)
 {
  switch(a)
  {
   case E_word:
    CheckRange(-32768L,65535L,&l);
    break;
   case E_adr:
    CheckRange(0L,65535L,&l);
    break;
   case E_byte:
    CheckRange(-128L,255L,&l);
    break;
   case E_ubyte:
    CheckRange(0L,255L,&l);
    break;
  }
  return(l);
 }
 else
 {
  if(l<E_syntax)
  {
   if(LenStr(par)>=max_explen)
   {
    Error(line,"expression '%s' is too long for forward reference",par);
   }
   else
   {
    CopyStr(par,fref[numfref].e);
    fref[numfref].lin=line;
    fref[numfref].adr=amem;
    fref[numfref].info=a;
    numfref++;
    if(numfref>=max_fref)
    {
     Error(line,"too many forward references (max %d)",max_fref);
     numfref=0;
    }
    if(numfref>maxnumfref)maxnumfref=numfref;
   }
  }
  if(a==E_branch)return(amem);
  else return(0L);
 }
}


void ClearLoc(void)
{
 ushort i,j=0;

 for(i=0;i<numlab;i++)
 {
  if(label[i].n[0]=='.')
  {
   if(!label[i].use)Warn(label[i].lin,"unused local symbol '%s'",label[i].n);
  }
  else
  {
   label[j].x=label[i].x;
   label[j].use=label[i].use;
   label[j].lin=label[i].lin;
   CopyStr(label[i].n,label[j].n);
   j++;
  }
 }
 numlab=j;
 for(i=0,j=0;i<numfref;i++)
 {
  if(CharInStr(fref[i].e,'.'))
  {
   Error(fref[i].lin,"undefined local symbol '%s'",fref[i].e);
  }
  else
  {
   fref[j].info=fref[i].info;
   fref[j].lin=fref[i].lin;
   fref[j].adr=fref[i].adr;
   CopyStr(fref[i].e,fref[j].e);
   j++;
  }
 }
 numfref=j;
}


void TryForward(void)
{
 ushort i,j=0;
 int l;

 for(i=0;i<numfref;i++)
 {
  if((l=Eval(fref[i].e))<E_FRef)
  {
   switch(fref[i].info)
   {
    case E_branch:
     l-=fref[i].adr+1L;
     if((l<-128L)||(l>127L))
     {
      if(l<0L)l=~l;
      l-=127L;
      Error(fref[i].lin,"branch out of range to line %d (%d bytes too far)",line,l);
     }
     PutMem(fref[i].adr,l&255L);
     break;
    case E_ubyte:
     CheckRange1(fref[i].lin,0L,255L,l);
     PutMem(fref[i].adr,l&255L);
     break;
    case E_adr:
     if((l>=0L)&&(l<256L)) Warn(fref[i].lin,"(line %d) generating 3-byte command although value is\n\t8-bit (zero page) due to forward reference", line);
     CheckRange1(fref[i].lin,0L,65535L,l);
     PutMem(fref[i].adr++,l&255L);
     PutMem(fref[i].adr,(l>>8L)&255L);
     break;
    case E_byte:
     CheckRange1(fref[i].lin,-128L,255L,l);
     PutMem(fref[i].adr,l&255L);
     break;
    case E_word:
     CheckRange1(fref[i].lin,-32768L,65535L,l);
     PutMem(fref[i].adr++,l&255L);
     PutMem(fref[i].adr,(l>>8L)&255L);
     break;
   }
  }
  else
  {
   fref[j].info=fref[i].info;
   fref[j].adr=fref[i].adr;
   fref[j].lin=fref[i].lin;
   CopyStr(fref[i].e,fref[j].e);
   j++;
  }
 }
 numfref=j;
}

void DefLabel(int l)
{
   ushort i;
   
   if((i=SearchLabel(lab))!=EOL) {
      Error(line,"symbol '%s' already defined in line %d",lab,label[i].lin);
      return;
   }
   if(LenStr(lab)>=max_lablen) {
      Warn(line,"symbol '%s' is too long and will be truncated to %d characters", lab, max_lablen-1);
      lab[max_lablen-1]=0;
   }
   if(lab[0]!='.')
     if(LenStr(lab)>pmaxlablen)
       pmaxlablen=LenStr(lab);
   label[numlab].x=l;
   label[numlab].use=false;
   label[numlab].lin=line;
   CopyStr(lab,label[numlab].n);
   numlab++;
   if(numlab>=max_lab) {
      Error(line,"too many symbols (max %d)",max_lab);
      numlab=0;
   }
   if(numlab>maxnumlab)
     maxnumlab=numlab;
   TryForward();
}

uchar mnldac[]={255,8,12,28,24,4,20,255,255,16,0};
uchar mnaslc[]={8,0,12,28,255,4,20,255,255,255,255};

//0=Inh,1=Imm,2=ABS,3=ABX,4=ABY,5=ZPG,6=ZPX,7=ZPY,8=IND,9=INY,10=INX
ushort GetAdrMode(void) {
   char *p;
   
   switch(par[0]) {
    case 0:
      return(0);
      
    case 'a':
    case 'A':
      if(par[1]==0)return(0);
      else goto def;
      
    case '#':
      CopyStr(par+1,par);
      return(1);
      
    case '(':
      p=par+LenStr(par)-3;
      if(p<par) p=par;
      if(SameStr(p,"),Y")||SameStr(p,"),y")) {
	 *p=0;
	 CopyStr(par+1,par);
	 return(9);
      }
      if(SameStr(p,",X)")||SameStr(p,",x)")) {
	 *p=0;
	 CopyStr(par+1,par);
	 return(10);
      }
      p=par+LenStr(par)-1;
      if(p<par) p=par;
      if(SameStr(p,")")) {
	 *p=0;
	 CopyStr(par+1,par);
	 return(8);
      }
      return(EOL);
      
      def: 
    default:
      if(!CharInStr(par,','))
	return(2);
      p=par+LenStr(par)-2;
      if(p<par) p=par;
      if(SameStr(p,",Y")||SameStr(p,",y")) {
	 *p=0;
	 return(4);
      }
      if(SameStr(p,",X")||SameStr(p,",x")) {
	 *p=0;
	 return(3);
      }
      return(EOL);
   }
}


char *mnemon="BPLBMIBVCBVSBCCBCSBNEBEQTXATAXTYATAYTSXTXSPHPPLP\
PHAPLACLCSECCLISEICLVCLDSEDDEYINYDEXINXBRKRTIRTS\
NOPJSRJMPORAANDEORADCSTALDACMPSBCASLROLLSRRORSTX\
LDXDECINCBITCPXCPYLDYSTY???BYTADRDSBDSAEQUORGEND\
ASSSETALGBECENC";


//0=Inh,f=branch,e=special,a=lda,b=ASL,c=cpx,d=pseudo,9=STY
ushort mnemonw[]={0xf10,0xf30,0xf50,0xf70,0xf90,0xfb0,0xfd0,0xff0,0x08a,0x0aa,0x098,0x0a8,0x0ba,0x09a,0x008,0x028,
      0x048,0x068,0x018,0x038,0x058,0x078,0x0b8,0x0d8,0x0f8,0x088,0x0c8,0x0ca,0x0e8,0x000,0x040,0x060,
      0x0ea,0xe20,0xe4c,0xa01,0xa21,0xa41,0xa61,0xa81,0xaa1,0xac1,0xae1,0xb02,0xb22,0xb42,0xb62,0x982,
      0xba2,0xbc2,0xbe2,0xc20,0xce0,0xcc0,0xba0,0x980,0x002,0xd00,0xd01,0xd02,0xd03,0xd04,0xd05,0xd06,
      0xd07,0xd08,0xd09,0xd0a,0xd0b};


ushort GetMNemonic(void) {
   char *m=mnemon;
   ushort i=0;
   
   if(opc[0]==0)
     return(EOL);
   if((opc[0]=='=')&&(opc[1]==0))
     return(0xd04);
   if(opc[3]!=0)return(ERR);
   while(*m) {
      if((opc[0]==m[0])&&(opc[1]==m[1])&&(opc[2]==m[2])) return(mnemonw[i]);
      i++;
      m+=3;
   }
   return(ERR);
}



ushort disvtab[]={0xfeff,0xffff,0x0400,0,0x5f04,0xffff,0xffff,0,
      0xffff,0xfeff,0x5f55,0,0x5d54,0xffff,0xfdff,0};


uchar GetMem(ushort a) {
   a -= base;
   if(a>=max_mem) return(0xbc);
   return mem[a] ;
}


char *dismtab="ORAANDEORADCSTALDACMPSBCASLROLLSRRORSTXLDXDECINC\
PHPCLCPLPSECPHACLIPLASEIDEYTYATAYCLVINYCLDINXSED\
BRKBPLJSRBMIRTIBVCRTSBVSLDXBCCLDYBCSCPYBNECPXBEQ\
???BITJMPJMPSTYLDYCPYCPXASLROLLSRRORTXATAXDEXNOPTXSTSX";

void DisAssO(uchar c)
{
 char *p=dismtab+c*3;

 par[0]=*p++;
 par[1]=*p++;
 par[2]=*p;
 par[3]=' ';
 par[4]=0;
}

ushort DisAssP(uchar c,ushort w)
{
 char *p=par+4;

 switch(c)
 {
  case 0://INX
   sprintf(p,"($%02X,X)",w&255);
   return(2);
  case 1://ZPG
   sprintf(p,"$%02X",w&255);
   return(2);
  case 2://IMM
   sprintf(p,"#$%02X",w&255);
   return(2);
  case 3://ABS
   sprintf(p,"$%04X",w);
   return(3);
  case 4://INY
   sprintf(p,"($%02X),Y",w&255);
   return(2);
  case 5://ZPX
   sprintf(p,"$%02X,X",w&255);
   return(2);
  case 6://ABY
   sprintf(p,"$%04X,Y",w);
   return(3);
  case 7://ABX
   sprintf(p,"$%04X,X",w);
   return(3);
  case 8://ZPY
   sprintf(p,"$%02X,Y",w&255);
   return(2);
  case 9://IND
   sprintf(p,"($%04X)",w);
 }
 return(3);
}


bool isIllegalCode(uchar c) {
   return (disvtab[c&15]&(1<<(c>>4)))==0; 
}


ushort DisAss(ushort a) {
   uchar c,b;

   c=GetMem(a++);
   if(!illegal) {
      sprintf(par, "BYT $%02X", c);
   } else {
      DisAssO(48);
   }   
   if(isIllegalCode(c)) return(1);
   if(c==0xa2)c=0x80;
   switch(c&15) {
    case 6:
    case 14:
      b=(c>>2)&7;
      if((c==0x96)||(c==0xb6))b=8;
      if(c==0xbe)b=6;
      DisAssO((c>>5)+8);
      return(DisAssP(b,GetMem(a++)+(GetMem(a)<<8)));
    case 1:
    case 5:
    case 9:
    case 13:
      DisAssO(c>>5);
      b=(c>>2)&7;
      return(DisAssP(b,GetMem(a++)+(GetMem(a)<<8)));
    case 8:
      DisAssO((c>>4)+16);
      return(1);
    case 0:
      DisAssO((c>>4)+32);
      if(c&0x10)
      {
	 b=GetMem(a++);
	 a+=(signed char)b;
	 DisAssP(3,a);
	 return(2);
      }
      else
      {
	 if(c==0x20) return(DisAssP(3,GetMem(a++)+(GetMem(a)<<8)));
	 if(c<128)return(1);
	 return(DisAssP(2,GetMem(a)));
      }
    case 10:
      if(c&0x10) DisAssO(((c>>5)&1)+64);
      else      DisAssO((c>>5)+56);
      return(1);
    default:
      DisAssO((c>>5)+48);
      b=(c>>2)&7;
      if(c==0x6c)b=9;
      return(DisAssP(b,GetMem(a++)+(GetMem(a)<<8)));
   }
}



/////////////////////////////////////////////////////////////////////////////
// interface functions

bool initAss(void) {
   int l;
   
   base=0;
   amem=0;
   while(GetLine()) {
      if(SameStr("ASS",opc)) {
	 bline=line;
	 l=Eval(par);
	 if(l>=E_FRef) {
	    ErrorRef();
	    l=0;
	 }
	 CheckRange(0L, 65535L, &l);
	 if(lab[0]) DefLabel(l);
	 base=l;
	 amem=l;
	 return true;
      }
   }
   return false;
}


void hexDump(int width) {
   int i=base, j;
   uchar b;

   par[width]=0;
   printf("hexdump of %d bytes (%dk), $%04X-$%04X:\n", amem-base, (amem-base+512)>>10, base, amem-1);
   while(i<amem) {
      printf(">%04X",i);
      for(j=0; j < width; j++, i++) {
	 b = GetMem(i);
	 if(j==8) printf(" - ");
	 else printf(" ");
	 if(i<amem) printf("%02X",b); 
	 else printf("  ");
	 par[j]='.';
	 if(isprint(b)) par[j]=b;
	 else par[j]='.';
	 if(i>=amem) par[j]=' ';
      }
      printf(" :%s\n",par);
   }
}


void disassemble(bool verbatim) {
   int i=base;
   bool print_adr = true;

      if(dis_source) {
	 printf("\n\tASS $%04x ; (last=$%04x, len=$%04x)\n\n", base, amem-1, amem-base);
      } else {
	 if(!verbatim) {
	    printf("disassembling %d bytes (%dk) of 6510-code, $%04X-$%04X:\n", amem-base, (amem-base+512)>>10, base, amem-1);
	 }
      }
   
   while(i<amem) {
      
      
      // DSB/DSA detect
      if((!no_dsb) && GetMem(i)==0) {
	 int j, len=0;
	 for(j=i; j<amem; j++, len++) 
	   if(GetMem(j)!=0) break;
	 if(len>=2) {
	    if(dis_source) {
	       printf("\nl_%04X\tDSB %5d\n", i, len);
	       print_adr = true;
	    } else printf(". %04X--%04X     DSB %5d\n", i, i+len-1, len);
	    i+=len;
	    continue;
	 }
      }
      
      // NOP field detect
      if((!no_nop) && GetMem(i)==0xea) {
	 int j, len=0;
	 for(j=i; j<amem; j++, len++) 
	   if(GetMem(j)!=0xea) break;
	 if(len>=2) {
	    if(dis_source) {
	       printf("\nl_%04X\tORG $%04X ; (%d x NOP)\n", i, i+len, len);
	       print_adr = true;
	    } else printf(". %04X--%04X     ORG $%04X ; (%d x NOP)\n", i, i+len-1, i+len, len);
	    i+=len;
	    continue;
	 }
      }
      
      // Illegal sequence detect
      if((!illegal) && isIllegalCode(GetMem(i))) {
	 int j, len=0;
	 for(j=i; j<amem; j++, len++) 
	   if(!isIllegalCode(GetMem(j))) break;
	 if(len>=2) {
	    do {
	       int l = len;
	       if(l>8) l = 8;
	       if(dis_source) {
		  printf("\nl_%04X\tBYT ", i);
		  print_adr = true;
	       } else printf(". %04X--%04X     BYT ", i, i+l-1);
	       for(j=0; j<l-1; j++) 
		 printf("$%02X,", GetMem(i+j));
	       printf("$%02X\n", GetMem(i+j));
	       j++;
	       i+=j;
	       len-=j;
	    } while(len>0);
	    continue;
	 }	 
      }

      // print adr
      if(dis_source && print_adr) {
	 printf("\nl_%04X", i);
	 print_adr = false;
      }
      
      // disassemble opcode
      int len;
      switch(DisAss(i)) {
       case 1:
	 if(!dis_source) printf(". %04X  %02X       %s\n", i, GetMem(i), par);
	 else printf("\t%s\n", par);
	 len=1;
	 break;
       case 2:
	 if(!dis_source) printf(". %04X  %02X %02X    %s\n", i, GetMem(i), GetMem(i+1), par);
	 else printf("\t%s\n", par);
	 len=2;
	 break;
       case 3:
	 if(!dis_source) printf(". %04X  %02X %02X %02X %s\n", i, GetMem(i), GetMem(i+1), GetMem(i+2), par);
	 else printf("\t%s\n", par);
	 len=3;
	 break;
       default:
	 fatalError("illegal case\n");
      }
      
      if(dis_source) {
	 switch(GetMem(i)) {
	  case 0x00: // BRK
	  case 0x60: // RTS
	  case 0x40: // RTI
	  case 0x4c: // JMP $xxxx
	  case 0x6c: // JMP ($xxxx)
	    print_adr = true;
	    break;
	    
	  default:
	    if(isIllegalCode(GetMem(i)))
	    print_adr = true;
	    break;
	 }
      }

      i+=len;
	
   }   
   if(dis_source) 
     printf("\n\nl_%04X\tEND\n\n", i);
}


void printUnused(void) {
   ushort i;
   int autolab = 0;
   
   for(i=0;i<numlab;i++) {
      if(!label[i].use) {
	 if((strlen(label[i].n)==6)&&(label[i].n[0]=='l')&&(label[i].n[1]=='_')) {
	    autolab++;
	 } else Warn(label[i].lin, "unused symbol '%s'", label[i].n);
      }
   }
   if(autolab) Warn(line, "%d unused auto symbols ('l_xxxx')", autolab);
   for(i=0;i<numfref;i++) {
      Error(fref[i].lin,"undefined symbol in expression '%s'", fref[i].e);
   }
}


void printStats(void) {
   double f;
   
   f=amem-base;
   printf("\n***** statistics *************************************************************\n");
   printf("* %d lines assembled into %d bytes (%.1fk) (avg. %.2f bytes per line)\n", line-bline, amem-base, f/1024.0, f/(line-bline));
   printf("* %d global symbols defined\n", numlab);
   switch(err) {
    case 0:printf("* no errors");
      break;
    case 1:printf("* one error");
      break;
    default:
      printf("* %d errors", err);
      break;
   }
   printf(" and ");
   switch(warn) {
    case 0:
      printf("no warnings");
      break;
    case 1:
      printf("one warning");
      break;
    default:
      printf("%d warnings", warn);
      break;
   }
   printf(" detected\n");
   printf("* internal memory usage:\n");
   f=amem-base;
   f/=max_mem/100.0;
   printf("* object code (bytes):%5d/%5d (%5.2f%%)\n", amem-base, max_mem, f);
   f=maxnumfref;
   f/=max_fref/100.0;
   printf("* forward references: %5d/%5d (%5.2f%%)\n", maxnumfref, max_fref, f);
   f=maxnumlab;
   f/=max_lab/100.0;
   printf("* symbols:            %5d/%5d (%5.2f%%)\n", maxnumlab, max_lab, f);
   printf("******************************************************************************\n");
}


void printSymbols(FILE *f) {
   ushort i,j;
   short sig;

   fprintf(f, "list of %d global symbols:\n", numlab);
   for(i=0; i<numlab; i++) {
      sig=j=(ushort)label[i].x;
      fprintf(f, "%-*s = $%04X == %5u == %6d defined on line %5d %s\n", pmaxlablen, label[i].n, j, j, sig, label[i].lin, label[i].use?"":"(unused)");
   }
}


bool assLine(void) {
   ushort m,a;
   int l;
   char *p=par;
   uchar c;

   if(!GetLine()) {
      Warn(line,"program should have an 'END' at the end");
      return(false);
   }
   m=GetMNemonic();

   if(m==0xd08) { //SET (needs to be before normal label treatment)
      l=Eval(par);
      if(l>=E_FRef) {
	 ErrorRef();
	 l=0;
      }
      if(lab[0]==0) {
	 Error(line, "'SET' needs symbol");
      } else {
	 a=SearchLabel(lab);
	 if(a==EOL) {
	    Warn(line, "'SET' creates symbol '%s'", lab);
	    DefLabel(l);
	 } else {
	    label[a].x=l;
	    label[a].lin=line;
	    label[a].use=false;
	 }
      }
      return(true);
   }
   
   if(m==0xd04) { //EQU/= (needs to be before normal label treatment)
      l=Eval(par);
      if(l>=E_FRef) {
	 ErrorRef();
	 l=0;
      }
      if(lab[0]) DefLabel(l);
      else {
	 Error(line,"'EQU/=' needs symbol");
      }
      return(true);
   } 
   
   // define normal label
   if(lab[0]) {
      DefLabel(amem);
      if(lab[0]!='.')ClearLoc();
   }
   
   // END/unknown/ASS/empty
   switch(m)
   {
    case 0xd06: // END
      ClearLoc();
      return(false);
    case EOL:   // empty line
      return(true);
    case ERR:   // unknown opcode
      Error(line,"unknown opcode %s",opc);
      return(true);
      
    case 0xd07: // ASS (is illegal here)
      Error(line,"'ASS' is allowed only once at the beginning");
      return(true);
   }
   
   // get address mode
   if((m>>8)!=0x0d)a=GetAdrMode();
   else a=0;
   if(a==EOL)
   {
      Error(line,"unknown address mode");
      return(true);
   }
   
   // assemble
   switch(m>>8)
   {
    case 0x0d://Pseudo opcode
      switch(m&255)
      {
       case 0://BYT
	 do {
	    p=par;
	    if(par[0]=='"') {
	       p++;
	       while((*p)&&(*p!='"'))PutMem(amem++,*p++);
	       if(*p==0)break;
	       p++;
	       if(*p==0)break;
	       if(*p==',')p++;
	       c=1;
	       CopyStr(p,par);
	       continue;
	    }
	    while((*p)&&(*p!=','))p++;
	    c=*p;
	    *p++=0;
	    if(par[0]==0) {
	       l=0;
	       Warn(line,"'BYT' without parameter inserts one zero byte ($00)");
	    }
	    else l=FEval(E_byte);
	    CheckRange(0,255,&l);
	    PutMem(amem++,l);
	    if(c)CopyStr(p,par);
	 }while(c);
	 break;
       case 1://ADR
	 do {
	    p=par;
	    while((*p)&&(*p!=','))p++;
	    c=*p;
	    *p++=0;
	    if(par[0]==0) {
	       l=0;
	       Warn(line,"'ADR' without parameter inserts one zero word ($0000)");
	    }
	    else l=FEval(E_word);
	    CheckRange(0,65535,&l);
	    PutMem(amem++,l&255L);
	    PutMem(amem++,(l>>8L)&255L);
	    if(c)CopyStr(p,par);
	 }while(c);
	 break;
       case 2://DSB define space byte
       case 3://DSA define space adr (word)
	 l=Eval(par);
	 if(l>=E_FRef) {
	    ErrorRef();
	    l=1;
	 }
	 CheckRange(0,16384,&l);
	 if((m&255)==3)l*=2;
	 while(l--)PutMem(amem++,0);
	 break;
       case 5://ORG set current pc
	 l=Eval(par);
	 if(l>=E_FRef) {
	    ErrorRef();
	    l=amem;
	 }
	 CheckRange(base,base+max_mem-1,&l);
	 if(l < amem) {
	    Warn(line, "'ORG' possibly overwrites %d bytes of code", amem - l);
	    amem = l;
	 } else {
	    while(amem < l) PutMem(amem++, 0xea); // put NOPs
	 }
	 break;
       case 9://ALG align
	 l=Eval(par);
	 if(l>=E_FRef) {
	    ErrorRef();
	    l=amem;
	 }
	 CheckRange(0,16,&l);
	 l = (1<<l)-1;
	 l = (amem+l)&(~l);
	 if(l<amem) fatalError("'ALG' miscalculation");
	 while(amem<l) PutMem(amem++, 0xea); // put NOPs
	 break;
       case 0xa://BEC begin comment
	 l = 1;
	 do {
	    if(!GetLine()) {
	       Warn(line,"end of file during 'BEC/ENC' comment");
	       return(false);
	    }
	    m=GetMNemonic();
	    if(m==0xd0a) l++; // BEC	       
	    if(m==0xd0b) l--; // ENC		 
	 } while(l > 0);
	 break;
       case 0xb://BEC begin comment
	 Error(line,"'ENC' (end of comment) without 'BEC' (begin of comment)"); // always an error
	 break;
      } // end of pseudo opcodes
      break;
      
    case 0x00://Inh
      PutMem(amem++,m&0xff);
      if(a!=0)Warn(line,"ignoring parameter for opcode '%s'", opc);
      break;
    case 0x0f://Branch
      PutMem(amem++,m&255);
      if(a!=2)Error(line,"banch needs absolute address mode");
      l=FEval(E_branch);
      l-=amem+1;
      if((l<-128L)||(l>127L))
      {
	 if(l<0L)l=~l;
	 l-=127L;
	 Error(line,"branch out of range (%d bytes to far)",l);
      }
      PutMem(amem++,l&255);
      break;
    case 0x0a://LDA...
      c=mnldac[a];
      if((c==255)||((a==1)&&((m&255)==0x81))) {
	 err:    Error(line,"invalid address mode");
	 break;
      }
      asl:   amem++;
      if(a) {
	 l=Eval(par);
	 if(l==E_syntax)goto e1;
      }
      if((a==2)||(a==3)) {
	 if(type==E_byte) {
	    a+=3;
	    c-=8;
	 }
      }
      switch(a) {
       case 0:break;
       case 2:
       case 3:
       case 4:
	 l=FEval(E_adr);
	 break;
       case 1:
	 l=FEval(E_byte);
	 break;
       default:
	 l=FEval(E_ubyte);
      }
      e1:   PutMem(amem-1,c|(m&255));
      if(a)PutMem(amem++,l&255L);
      switch(a) {
       case 2:
       case 3:
       case 4:
	 PutMem(amem++,(l>>8L)&255L);
      }
      break;
    case 0x0b://ASL...
      c=mnaslc[a];
      switch(m&255) {
       case 0x02:
       case 0x22:
       case 0x42:
       case 0x62:
	 if(c==0)goto err;
	 break;
       case 0xa2:
	 if(a==3)goto err;
	 if(a==4)
	 {
	    c=28;
	    a=3;
	 }
	 break;
       case 0xa0:
	 if(c==8)goto err;
	 break;
       case 0xc0:
       case 0xe0:
	 if((c==8)||(c==0))goto err;
	 break;
      }
      if(c==255)goto err;
      goto asl;
    case 0x0c://CPX...
      switch(a) {
       case 1:
	 if((m&255)==0x20)goto err;
	 c=0;
	 break;
       case 2:
	 c=12;
	 break;
       default: goto err;
      }
      goto asl;
    case 0x09://STX...
      if((m&255)==0x82) {
	 if(a==3)goto err;
	 if(a==4)a=3;
      }
      switch(a) {
       case 2:
	 c=12;
	 break;
       case 3:
	 a=6;
	 c=20;
	 break;
       default: goto err;
      }
      goto asl;
    case 0x0e://JSR,JMP
      switch(a) {
       case 8:
	 if((m&255)==0x20)goto err;
	 m=0x6c;
       case 2:
	 PutMem(amem++,m&255);
	 l=FEval(E_adr);
	 PutMem(amem++,l&255L);
	 PutMem(amem++,(l>>8L)&255L);
	 break;
       default: goto err;
      }
   }
   return(true);
}






